//##############################################################################
// The $class Library
// Copyright 2006, Jeff Lau
// License: http://creativecommons.org/licenses/LGPL/2.1/
// Contact: jlau@uselesspickles.com
// Web:     www.uselesspickles.com
//##############################################################################

/**
 * @fileOverview
 * <p>
 * This is the debug version of the $class library, which contains tons of useful error checking. Include this
 * file during development or debugging. When it's time to deploy, be sure to switch over to including class.js
 * to get rid of the error checking overhead.
 * </p>
 *
 * <p>
 * <h2>
 * Error/Warning Logging
 * </h2>
 *   The debug version of the $class library attempts to log warning  and error
 * messages by methods "console.warn" and "console.error". These are logging
 * methods supported by the Firebug plugin for the Firefox web browser. If you
 * are debugging in a different environment and would like to log warnings/errors
 * in some way, create a global named "console" with methods named "warn" and
 * "error". These methods should expect to receive a single argument and should
 * log/display the value in whatever way is appropriate for your logging needs.
 * The logging methods should be defined before any calls to the $class
 * library to ensure all warnings are logged.
 * 
 * Example:
<pre class="brush:js">
var console = {
  warn: function(message) {
    alert("Warning: " + message);
  },

  error: function(message) {
    alert("Error: " + message);
  }
};
</pre>
 * </p>
 */

/**
 * @namespace
 * This is where the $class library stores some of its implementation details to avoid polluting the global
 * scope.
 */
$class_library = {
  /**
   * The version of the $class library.
   * @type String
   */
  version: "2.0b",
  
  /**
   * True if this is the debug version of the library.
   * @type Boolean
   */
  debug: true,

  _globalObject: this,

  _hasOwnProperty: Object.prototype.hasOwnProperty ?
    function (obj, prop) {
      return obj.hasOwnProperty(prop);
    }
    :
    function (obj, prop) {
      return (prop in obj) && obj.constructor.prototype[prop] !== obj[prop];
    },

  _warn: function() {
    var global = this._globalObject;

    if (global.console && global.console.warn) {
      var apply = Function.prototype.apply;
      apply.call(global.console.warn, global.console, arguments);
    }
  },

  _error: function() {
    var global = this._globalObject;

    if (global.console && global.console.error) {
      var apply = Function.prototype.apply;
      apply.call(global.console.error, global.console, arguments);
    }

    var message = "";

    for (var i = 0; i < arguments.length; ++i) {
      if (i) {
        message += "; ";
      }

      message += arguments[i];
    }

    throw new Error(message);
  },

  _copyObject: function(obj1, obj2) {
    var result = (arguments.length == 2 && obj1) ? obj1 : {};
    var source = ((arguments.length == 2) ? obj2 : obj1) || {};

    for (var i in source) {
      result[i] = source[i];
    }

    return result;
  },

  _surrogateCtor: function() {
  }
};

//##############################################################################

/**
 * <p>
 * Creates a namespace. Note that this documentation describes the behavior of $namespace when used as a method
 * (without the "new" keyword). This should only be called as a method rather than as a constructor. Instances 
 * of $namespace are created as a private implementation detail of $namespace being called as a method.
 * </p>
 * <p>
 * A namespace is simply a named "container" for classes, methods, properties or even  other namespaces. 
 * </p>
 *
 * <p>
 * <h2>Example:</h2>
<pre class="brush:js">
$namespace("MyLibrary.UI.Forms");

MyLibrary.UI.Forms.VERSION = 5;

$class("MyLibrary.UI.Forms.Field", {
  // ... properties, methods, etc.
});

var field = new MyLibrary.UI.Forms.Field();
</pre>
 * </p>
 *
 * @class 
 * <p>
 * This class represents a namespace, but its constructor is used as a function (without the "new" 
 * keyword) to create a namespace. It is used as a constructor for private implementation purposes only. 
 * </p>
 *
 * @constructor
 *
 * @param {String} name The name of the namespace. The name may describe a path of multiple nested 
 * namespace/class/object names separated  by a period (.). All namespaces along the path will be created as 
 * necessary. For example, a namespace name of "MyLibrary.UI.Forms" will create a namespace named "MyLibrary"
 * that contains a namespace named "UI" which in turn contains a namespace named "Forms". If any of the 
 * namespaces in the specified name path already exist, but were not created by $namespace, a warning will be 
 * issued by the debug version of the $class library.
 *
 * @return {$namespace} the newly created (or pre-existing) namespace. If an entire path of namespaces was 
 * created, only the last namespace in the path is returned.
 */
function $namespace(name) {
  // if being called as a constructor...
  if (this instanceof $namespace) {
    /**
      * Overridden to return a string that identifies this namespace. This may be useful for debugging purposes. 
      * This also gives each namespace a unique
      * string representation for use as object keys.
      * @return {String} a string in the format "[$namespace _name_]", where _name_ is the name of the 
      * namespace, including the full path of parent namespaces.
      */
    this.toString = function() {
       return "[$namespace " + name + "]";
    };
    
    $class_library._surrogateCtor.prototype = this;
    return new $class_library._surrogateCtor();
  }

  var components = name.split(".");
  var context = $class_library._globalObject;

  for (var i = 0; i < components.length; ++i) {
    var nextContext = context[components[i]];

    if (!nextContext) {
      nextContext = new $namespace(components.slice(0, i + 1).join("."));
      context[components[i]] = nextContext;
    } else if (!(nextContext instanceof $namespace)) {
      $class_library._warn("Warning in $namespace(\"" + name + "\"): [" +
                            components.slice(0, i + 1).join(".") +
                            "] is already defined, but is not a namespace.");
    }

    context = nextContext;
  }

  return context;
}

//##############################################################################

/**
 * <p>
 * Creates a class with the specified name and properties. Note that this documentation describes the behavior 
 * of $class when used as a method (without the "new" keyword). This should only be called as a method rather 
 * than as a constructor. Instances of $class are created as a private implementation detail of $class being 
 * called as a method.
 * </p>
 *
 * <p>
 * $class supports many features, such as abstract methods, implementing interfaces, easily calling an 
 * overridden implementation of a method, and much more. See {@link $class.descriptor} for details.
 * </p> 
 *
 * <p>
 * The new class will have a static property named "$class" that is an instance of $class and can provide
 * information about the class that was created by $class. Should I say "class" a few more times?
 * </p> 
 *
 * <p>
 * <h2>Example:</h2>
<pre class="brush:js">
$class("Sample", {
  $constructor: function(arg) {
    this._property = arg;
  },
  
  getProperty: function() {
    return this._property;
  }
});

var sample = new Sample("Hello World!");

alert(sample.getProperty()); // alerts "Hello World"

alert(sample instanceof Sample); // alerts "true"
</pre>
 * </p>
 *
 * @class This class encapsulates information about a class, but its constructor is used as a function 
 * (without the "new" keyword) to create a class. It is used as a constructor for private implementation 
 * purposes only. Each class created by $class has a static property named "$class" that is an instance 
 * of $class. $class instances can also be obtained by calling {@link #getClass}.
 *
 * @constructor
 *
 * @param {String} name The name of the class. The name may describe a path of multiple nested 
 * namespace/class/object names separated by a period (.). The entire path to the class name is expected 
 * to exist already. See {@link $namespace} for details about creating namespaces. 
 *
 * @param {$class.descriptor} descriptor An object containing the properties of the class, which can describe 
 * methods, static methods, static properties, and more.
 *
 * @return {Function} The constructor for the newly created class.
 */
function $class(name, descriptor) {
  // if being called as a constructor...
  if (this instanceof $class) {
    if (!descriptor.calledFrom$class && !descriptor.calledFrom$class_adapt) {
      $class._error(name, "Do not directly instantiate $class! Call $class as a function instead.");
    }

    this._name = name;
    this._isNative = descriptor.isNative;
    this._ctor = descriptor.ctor;
    this._baseCtor = descriptor.baseCtor || (this._ctor == Object ? null : Object);
    this._isFinal = Boolean(descriptor.isFinal);
    this._isSingleton = Boolean(descriptor.isSingleton);
    this._isMultiton = Boolean(descriptor.isMultiton);
    this._isAbstract = false;
    this._interfaces = descriptor.interfaces || {};
    this._abstractMethods = {};
    this._finalMethods = {};
    this._interfaceMethods = {};
    this._children = [];

    // if not creating $class for Object...
    if (this._ctor != Object) {
      // if inheriting something other than Object...
      if (this._baseCtor != Object) {
        // if this $class object is being created internally by the $class function...
        if (descriptor.calledFrom$class) {
          // a 'surrogate' constructor used to create inheritance relationship
          // without actually invoking the base class's constructor code
          $class_library._surrogateCtor.prototype = this._baseCtor.prototype;
          this._ctor.prototype = new $class_library._surrogateCtor();
          this._ctor.prototype.constructor = this._ctor;
        }

        // inherit info about the base class
        var base$class = this._baseCtor.$class;

        if (base$class) {
          this._interfaces = $class_library._copyObject(base$class._interfaces);
          this._abstractMethods = $class_library._copyObject(base$class._abstractMethods);
          this._finalMethods = $class_library._copyObject(base$class._finalMethods);

          base$class._children.push(this);
        } else {
          this._interfaces = {};
          this._abstractMethods = {};
          this._finalMethods = {};
        }
      }

      // store this class info on the prototype
      this._ctor.prototype._$class = this;
    }

    // Storing class info on prototype and constructor causes an IE leak; store created $class objects to clean up when
    // unloading the page.
    $class._cleanupCache.push(this);
    return;
  }

  if ($class_library._hasOwnProperty(descriptor, "$extends") && !descriptor.$extends) {
    $class._propertyError(name, "$extends", "Cannot extend null. You probably forgot to include the JS file containing the declaration of the parent class.");
  }

  var baseCtor = descriptor.$extends || Object;
  var isSingleton = !!descriptor.$singleton;
  var isMultiton = !!descriptor.$multiton;

  if (!(baseCtor instanceof Function)) {
    $class._propertyError(name, "$extends", "Must be a constructor function");
  }

  if (!baseCtor.$class) {
    $class._propertyWarn(name, "$extends", "Extending a class that has no $class information. "
                                         + "The class has been automatically adapted and given the name "
                                         + "\"" + name + "$base\". Use $class.adapt() to pre-adapt "
                                         + "the class and provide the correct name");
    $class.adapt(baseCtor, name + "$base");
  }

  if (baseCtor.$class && baseCtor.$class._isFinal) {
    $class._propertyError(name, "$extends", "Cannot extend [" + baseCtor.$class.getName() + "] because its constructor is declared $final.");
  }

  if (!isSingleton && baseCtor.$class && baseCtor.$class._isSingleton) {
    $class._propertyError(name, "$extends", "Cannot extend [" + baseCtor.$class.getName() + "] because it is a singleton. Singletons can only be extended by other singletons.");
  }

  if (!isMultiton && baseCtor.$class && baseCtor.$class._isMultiton) {
    $class._propertyError(name, "$extends", "Cannot extend [" + baseCtor.$class.getName() + "] because it is a multiton. Multitons can only be extended by other multitons.");
  }

  if (isSingleton && isMultiton) {
    $class._propertyError(name, "$singleton/$multiton", "Cannot be both a singleton and a multiton");
  }

  var ctorName = name.replace(/[^.]*\./g, "");
  var ctor = descriptor.$constructor || descriptor[ctorName];
  var isFinal = false;

  if (ctor instanceof $class._ModifiedProperty && ctor.getModifier() == "final") {
    ctor = ctor.getValue();
    isFinal = true;
  }

  if (!ctor) {
    $class_library._warn("No constructor specified for class [" + name + "]; using an empty default constructor.");
    ctor = new Function();
  }

  if (!(ctor instanceof Function)) {
    $class._propertyError("$constructor", "Must be a function");
  }

  ctor = $class._wrapCtorMethod(ctor, baseCtor);
  ctor.$class = new $class(name, {ctor:ctor, baseCtor:baseCtor, isFinal:isFinal, isSingleton:isSingleton, isMultiton:isMultiton, calledFrom$class:true});

  if (isSingleton) {
    var createInstance = (descriptor.$singleton.createInstance instanceof Function) ? descriptor.$singleton.createInstance : $class._createSingletonInstance_implementation;

    descriptor._$class_createSingletonInstance = $static(createInstance);
    descriptor._$class_singletonInstance = $static(null);
    descriptor.getInstance = $static($class._getSingletonInstance_implementation);
  }

  if (isMultiton) {
    var createInstance = (descriptor.$multiton.createInstance instanceof Function) ? descriptor.$multiton.createInstance : $class._createMultitonInstance_implementation;
    var createCacheKey = (descriptor.$multiton.createCacheKey instanceof Function) ? descriptor.$multiton.createCacheKey : $class._createMultitonCacheKey_implementation;

    descriptor._$class_createMultitonInstance = $static(createInstance);
    descriptor._$class_createMultitonCacheKey = $static(createCacheKey);
    descriptor._$class_multitonInstances = $static({});
    descriptor._$class_creatingMultitonInstance = $static({});
    descriptor.getInstance = $static($class._getMultitonInstance_implementation);
  }

  // implement interfaces
  if (descriptor.$implements != null) {
    var ifaces = descriptor.$implements;

    // convert to an array if it is a single object
    if (!(ifaces instanceof Array)) {
      ifaces = [ifaces];
    }

    for (var i = 0, ifacesLength = ifaces.length; i < ifacesLength; ++i) {
      // make sure the 'interface' to extend is really an interface
      if (!(ifaces[i] instanceof $interface)) {
        $class._propertyError(name, "$implements", "$interface or array of $interfaces expected");
      }

      ctor.$class._implement(ifaces[i]);
    }
  }

  var specialProperties = {$constructor:true,$singleton:true,$multiton:true,$extends:true,$implements:true,$static:true};
  var processedProperties = {};
  specialProperties[ctorName] = true;

  // process all properties in the class descriptor
  for (var propertyName in descriptor) {
    // skip over special properties
    if ($class_library._hasOwnProperty(specialProperties, propertyName)) {
      continue;
    }

    ctor.$class._processProperty(propertyName, descriptor[propertyName]);
    processedProperties[propertyName] = true;
  }

  if (!$class_library._hasOwnProperty(processedProperties, "toString") && $class_library._hasOwnProperty(descriptor, "toString")) {
    ctor.$class._processProperty("toString", descriptor["toString"]);
  }

  if (!$class_library._hasOwnProperty(processedProperties, "valueOf") && $class_library._hasOwnProperty(descriptor, "valueOf")) {
    ctor.$class._processProperty("valueOf", descriptor["valueOf"]);
  }

  // collect names of all interface methods that are not implemented
  var missingInterfaceMethods = ctor.$class._getMissingInterfaceMethods();

  // if any interface methods are not implemented, we have a problem
  if (missingInterfaceMethods.length != 0) {
    var message = "The following interface methods must be implemented: ";

    for (var i = 0; i < missingInterfaceMethods.length; ++i) {
      message += "[" + missingInterfaceMethods[i] + "] ";
    }

    $class._error(name, message);
  }

  // if any abstract methods are remaining, this class is abstract
  for (var methodName in ctor.$class._abstractMethods) {
    ctor.$class._isAbstract = true;
    break;
  }

  // set default toString method
  if (ctor.prototype.toString == Object.prototype.toString) {
    ctor.prototype.toString = new Function(
      "return \"[object \" + $class.typeOf(this) + \"]\";"
    );
  }

  // store the constructor so it is accessible by the specified name
  try {
    var result = eval(name + " = ctor;");
  } catch (error) {
    $class._error(name, "Invalid class name: " + error.message);
  }

  // call the static initializer
  if (descriptor.$static instanceof Function) {
    try {
      descriptor.$static.call(ctor);
    } catch (error) {
      $class._error(name, "Error while executing static initializer: " + error.message);
    }
  } else if (descriptor.$static != null) {
    $class._propertyError(name, "$static", "function expected");
  }

  return result;
}

//##############################################################################

$class.prototype = {
  //############################################################################

  /**
   * Gets the full name of the class.
   * @return {String} the full name of the class.
   */
  getName: function() {
    return this._name;
  },

  //############################################################################

  /**
   * Returns true if this class is one of the standard JavaScript classes (Number, Function, etc).
   * @return {Boolean} true if this class is one of the standard JavaScript classes.
   */
  isNative: function() {
    return this._isNative;
  },

  //############################################################################

  /**
   * Returns the prototype object of this class.
   * @return {Object} The class's prototype object.
   */
  getPrototype: function() {
    return this._ctor.prototype;
  },

  //############################################################################

  /**
   * Gets the constructor for the class.
   * @return {Function} the constructor for the class.
   */
  getConstructor: function() {
    return this._ctor;
  },

  //############################################################################

  /**
   * Gets the $class object of this class's parent class.
   * @return {$class} The $class object for this class's parent class. Returns null if this class is Object.
   */
  getSuperclass: function() {
    return this._baseCtor ? this._baseCtor.$class : null;
  },

  //############################################################################

  /**
   * Tests if an object is an instance of this class.
   * @param {Object} obj
   * @return {Boolean} true if the object is an instance of this class, or any of its child classes.
   */
  isInstance: function(obj) {
    return $class.instanceOf(obj, this._ctor);
  },

  //############################################################################

  /**
   * Tests if this class implements a specific interface.
   * @param {$interface} iface
   * @return {Boolean} true if this class implements the interface.
   */
  implementsInterface: function(iface) {
    return $class_library._hasOwnProperty(this._interfaces, iface.getName());
  },

  //############################################################################

  /**
   * Overridden to return a string that identifies this $class object.
   * @return {String} a string in the form "[$class _name_]", where _name_ is the full name of this class.
   */
  toString: function() {
    return "[$class " + this._name + "]";
  },

  //############################################################################

  _implement: function(iface) {
    $class_library._copyObject(this._interfaces, iface._interfaces);
    $class_library._copyObject(this._interfaceMethods, iface._methods);
  },

  //############################################################################

  _processProperty: function(propertyName, value) {
    var modifier = "";

    if (value instanceof $class._ModifiedProperty) {
      modifier = value.getModifier();
      value = value.getValue();

      if (value instanceof $class._ModifiedProperty) {
        $class._propertyError(name, propertyName, "Only one modifier may be used per value");
      }

      switch (modifier) {
        case "static":
          this._ctor[propertyName] = value;
          return;

        case "abstract":
        case "final":
          if (!(value instanceof Function)) {
            $class._propertyError(name, propertyName, "The $" + modifier + " modifier may only be applied to function values.");
          }
          break;
      }
    }

    // don't override anything that is final
    if ($class_library._hasOwnProperty(this._finalMethods, propertyName)) {
      var className = this._finalMethods[propertyName];
      $class._propertyError(this._name, propertyName, "Cannot override a final property (originally declared final in class [" + className + "])");
    }

    var isAbstract = false;

    switch (modifier) {
      case "final":
        this._finalMethods[propertyName] = this._name;
        break;

      case "abstract":
        this._abstractMethods[propertyName] = this._name;
        value = $class._createAbstractMethod(this._name, propertyName);
        isAbstract = true;
        break;
    }

    // if the property is overriding one from an inherited class...
    if (propertyName in this._ctor.prototype) {
      var baseValue = this._ctor.prototype[propertyName];

      // only allow functions to be overridden
      if (!(baseValue instanceof Function) || !(value instanceof Function)) {
        $class._propertyError(this._name, propertyName, "Only function properties can be overridden (with another function)");
      } else {
        if ($class._uses$base.test(value) && !$class_library._hasOwnProperty(this._abstractMethods, propertyName)) {
          // wrap the method to give it access to the special $base property
          value = $class._wrapExtendedMethod(value, baseValue);
        }
      }

      if (!isAbstract) {
        delete this._abstractMethods[propertyName];
      }
    }

    this._ctor.prototype[propertyName] = value;
  },

  //############################################################################

  _getMissingInterfaceMethods: function() {
    var result = new Array();

    for (var methodName in this._interfaceMethods) {
      if (!(this._ctor.prototype[methodName] instanceof Function)) {
        var ifaceName = this._interfaceMethods[methodName];
        result.push(ifaceName + "." + methodName);
      }
    }

    // don't need this info any more
    delete this._interfaceMethods;

    return result;
  }

  //############################################################################
};

//##############################################################################
// List of $class objects created so they can be cleaned up when unloading the page
$class._cleanupCache = [];

// using attachEvent because this is only needed to avoid memory leaks in IE, and only IE supports attachEvent.
if (typeof window != "undefined" && window.attachEvent) {
  window.attachEvent('onunload', function() {
    for (var i=0,clz; clz = $class._cleanupCache[i]; i++) {
      delete clz._ctor.prototype._$class;
      delete clz._ctor.$class;
    }
  });
}

//##############################################################################

$class._uses$base = /\bthis\.\$base\b/;

//##############################################################################

$class._error = function(name, message) {
  $class_library._error("Error in $class(\"" + name + "\"" + ", ...): " + message);
};

//##############################################################################

$class._propertyError = function(name, propertyName, message) {
  $class_library._error("Error in $class(\"" + name + "\"" + ", ...), property [" + propertyName + "]: " + message);
};

//##############################################################################

$class._propertyWarn = function(name, propertyName, message) {
  $class_library._warn("Warning in $class(\"" + name + "\"" + ", ...), property [" + propertyName + "]: " + message);
};

//##############################################################################

$class._wrapCtorMethod = function(method, baseMethod) {
  // automatically call the base class constructor if inheriting something
  // other than Object and the constructor does not already call it
  var call$base = !$class._uses$base.test(method) && baseMethod != Object;
  var result = $class._createCtorWrapper(call$base);

  result._$class_wrappedMethod = method;
  result.toString = $class._wrappedMethod_toString;

  return result;
};

//##############################################################################

$class._createCtorWrapper = function(call$base) {
  return function() {
    var method = arguments.callee;
    var ctor = $class.getClass(this).getConstructor();

    // if this is the actual constructor for the object (not a base class constructor)...
    if (ctor == method) {
      if (method.$class._isAbstract) {
        var message = "Attempted instantiation of the abstract class [" +
                      method.$class._name + "]. Abstract methods: ";

        for (var methodName in method.$class._abstractMethods) {
          message += "[" + method.$class._abstractMethods[methodName] + "." +
                     methodName + "] ";
        }

        $class_library._error(message);
      }

      if (ctor.$class._isSingleton && !ctor._$class_creatingSingletonInstance) {
        var message = "Attempted instantiation of the singleton class [" +
                      ctor.$class._name + "]. Use " + ctor.$class._name + ".getInstance()";

        $class_library._error(message);
      }

      if (ctor.$class._isMultiton && !ctor._$class_creatingMultitonInstance[ctor._$class_createMultitonCacheKey.apply(ctor, arguments)]) {
        var message = "Attempted instantiation of the multiton class [" +
                      ctor.$class._name + "]. Use " + ctor.$class._name + ".getInstance([constructor args])";

        $class_library._error(message);
      }
    }

    var previousBase = this.$base;
    this.$base = method.$class._baseCtor;

    try {
      if (call$base) {
        this.$base.apply(this, arguments);
      }

      return method._$class_wrappedMethod.apply(this, arguments);
    } finally {
      this.$base = previousBase;
    }
  };
};

//##############################################################################

$class._wrapExtendedMethod = function(method, baseMethod) {
  var result = $class._createExtendedMethodWrapper();

  result._$class_wrappedMethod = method;
  result._$class_baseMethod = baseMethod;
  result.toString = $class._wrappedMethod_toString;

  return result;
};

//##############################################################################

$class._createExtendedMethodWrapper = function() {
  return function() {
    var method = arguments.callee;

    var previousBase = this.$base;
    this.$base = method._$class_baseMethod;

    try {
      return method._$class_wrappedMethod.apply(this, arguments);
    } finally {
      this.$base = previousBase;
    }
  };
};

//##############################################################################

$class._wrappedMethod_toString = function() {
  return this._$class_wrappedMethod.toString();
};

//##############################################################################

$class._createAbstractMethod = function(name, propertyName) {
  return function() {
    $class_library._error("The abstract method [" + propertyName + "] declared " +
                          "by class [" + name + "] was invoked on an object of type [" +
                          $class.typeOf(this) + "].");
  };
};

//##############################################################################

$class._getSingletonInstance_implementation = function() {
  if (this._$class_singletonInstance) {
    return this._$class_singletonInstance;
  }

  if (this._$class_creatingSingletonInstance) {
    $class_library._error(this.$class.getName() + ".getInstance(): infinite recursion encountered. Check for a call to getInstance that is called as a result of instantiating this class.");
  }

  this._$class_creatingSingletonInstance = true;
  this._$class_singletonInstance = this._$class_createSingletonInstance();
  delete this._$class_creatingSingletonInstance;

  return this._$class_singletonInstance;
};

//##############################################################################

$class._createSingletonInstance_implementation = function() {
  return new this();
};

//##############################################################################

$class._getMultitonInstance_implementation = function() {
  var cacheKey = this._$class_createMultitonCacheKey.apply(this, arguments);
  var result = this._$class_multitonInstances[cacheKey];

  if (result) {
    return result;
  }

  if ($class_library._hasOwnProperty(this._$class_creatingMultitonInstance, cacheKey)) {
    $class_library._error(this.$class.getName() + ".getInstance(): infinite recursion encountered. Check for a call to getInstance that is called as a result of instantiating this class.");
  }

  this._$class_creatingMultitonInstance[cacheKey] = true;
  result = this._$class_multitonInstances[cacheKey] = this._$class_createMultitonInstance.apply(this, arguments);
  delete this._$class_creatingMultitonInstance[cacheKey];

  return result;
};

//##############################################################################

$class._createMultitonInstance_implementation = function() {
  return $class.instantiate(this, arguments);
};

//##############################################################################

$class._createMultitonCacheKey_implementation = function() {
  return Array.prototype.join.call(arguments, "|");
};

//##############################################################################

$class._ModifiedProperty = function(modifier, value) {
  this._modifier = modifier;
  this._value = value;
};

$class._ModifiedProperty.prototype = {
  getModifier: function() {
    return this._modifier;
  },

  getValue: function() {
    return this._value;
  }
};

//##############################################################################

/**
 * <p>
 * Adapts a pre-existing class into the $class library framework. Use this to give a pre-existing class a 
 * {@link $class} object so that it can fully participate in $class library features. For example, retroactively
 * applied interfaces ({@link $interface#applyInterface}) will be automatically applied to all descendents
 * of a class only if that class was defined by the $class function, or if the class was adapted with this method.
 * </p>
 *
 * <p>
 * If the specified class extends another class that does not exist within the $class framework, then that parent
 * class will be automatically adapted, but with a name that is simply the specified name with "$base" appended
 * (there's no way for me to know the actual name of the parent class). If you want each class in the heirarchy
 * to have the correct name, then you must adapt each class, starting from the base class.
 * </p>
 *
 * <p>
 * Note that it is not necessary to adapt a class if you simply want to extend it using the $class library. When
 * extending a class, it will be automatically adapted with a name made up of the child class's name with "$base"
 * appended.
 * </p>
 *
 * <p>
 * If any {@link $interface}s had been applied ({@link $interface#applyInterface}) to the class or its prototype 
 * before being adapted, it will be detected when adapting the class so that the adapted class and all of its 
 * decendents will be considered to implement the interfaces.
 * </p>
 *
 * <p>
 * <h2>Example:</h2>
<pre class="brush:js">
var MyNamespace = {};

MyNamespace.Sample = function(param) {
  this._param = param;
}

MyNamespace.Sample.prototype = {
  getParam: function() {
    return 
  }
};

$class.adapt(MyNamespace.Sample, "MyNamespace.Sample");

alert(MyNamespace.Sample.$class.getName()); // alerts "MyNamespace.Sample"
</pre>
 * </p>
 *
 * <p>
 * <h2>Debug version warnings:</h2>
 * <ul>
 *   <li>No name is specified; an anonymous name will be supplied.</li>
 *   <li>Base class of the adapted class is automatically adapted.</li>
 * </ul>
 * </p>
 *
 * @function
 * @param {Function} ctor The constructor function of the class to be adapted.
 * @param {String} name (optional) The full name of the class. If not specified, an anonymous name will be 
 * generated.
 * @return {Boolean} true if the class was adapted, false if the class is already within the $class library 
 * framework at the time this method was called.
 */
$class.adapt = (function() {
  var unnamedCount = 0;

  return function(ctor, name, isNative) {
    if (ctor.$class) {
      return false;
    }

    var baseCtor = ctor.prototype.constructor;

    if (ctor != Object) {
      var hold = ctor.prototype.constructor;

      if (delete ctor.prototype.constructor) {
        baseCtor = ctor.prototype.constructor;
        ctor.prototype.constructor = hold;
      }
    }

    if (!name) {
      name = "$unnamed_adapted_class_" + (unnamedCount++);
      
      $class_library._warn("Warning in $class.adapt([function]): "
                         + "No name was specified. "
                         + "Defaulting the name to \"" + name + "\".");      
    }

    if (baseCtor == ctor || !baseCtor) {
      baseCtor = Object;
    } else if (!baseCtor.$class) {
      $class_library._warn("Warning in $class.adapt([function], \"" + name + "\"): "
                         + "The class being adapted extends a class that must also be adapted. "
                         + "It will be given the name \"" + name + "$base\". "
                         + "Pre-adapt the base class to give it the correct name.");
      $class.adapt(baseCtor, name + "$base");
    }

    var interfaces = ctor.prototype._$class_interfaces;
    delete ctor.prototype._$class_interfaces;
    
    ctor.$class = new $class(name, {
      ctor: ctor,
      baseCtor: baseCtor,
      isNative: isNative,
      interfaces: interfaces,
      calledFrom$class_adapt: true
    });

    return true;
  };
})();

//##############################################################################

/**
 * <p>
 * Gets the value of a fully qualified name, possibly nested within one or more {@link $namespace}s or
 * namespace-like objects. Does not cause an error if any object along the path is undefined/null. Simply
 * returns undefined of the specified name could not be reached.
 * </p>
 *
 * <p>
 * <h2>Example:</h2>
<pre class="brush:js">
$namespace("nested.namespace");

nested.namespaces.SOME_VALUE = 10;

alert($class.resolve("nested.namespace.SOME_VALUE")) // alerts "10"

alert($class.resolve("nonexistant.namespace.SOME_VALUE")) // alerts "undefined"
</pre>
 * </p>
 *
 * @param {String} name The fully qualified name of some object/value.
 * @return {mixed} the value found at the specified name, or undefined if it could not be reached.
 */
$class.resolve = function(name) {
  var components = name.split(".");
  var context = $class_library._globalObject;

  for (var i = 0; i < components.length; ++i) {
    if (context == null) {
    $class_library._warn("$class.resolve(\"" + name + "\") failed: "
                        + components.slice(0, i).join(".") + " is null/undefined.");
      return undefined;
    }

    var context = context[components[i]];
  }

  return context;
};

//##############################################################################

/**
 * <p>
 * Tests if an object is an implementation of an {@link $interface}. This method only tests if the object has 
 * already been marked as being an implementation of the interface; it does not actually analyze the object to
 * determine if all interface methods are implemented. See {@link $class.descriptor#$implements} and 
 * {@link $interface#applyInterface} for performing the work of marking an object as implementing an interface.
 * </p>
 *
 * @param {mixed} obj any object or primitive.
 * @param {$interface} iface any interface.
 * @return {Boolean} true if the object implements the interface.
 */
$class.implementationOf = function(obj, iface) {
  if (obj._$class_interfaces && $class_library._hasOwnProperty(obj._$class_interfaces, iface.getName())) {
    return true;
  }

  return $class.getClass(obj).implementsInterface(iface);
};

//##############################################################################

/**
 * <p>
 * Tests if an object is an instance of a class or an implementation of an interface.
 * </p>
 *
 * <p>
 * This method has special handling of primitive JavaScript values to consider them to be instances of their
 * corresponding classes. For example, <code class="brush:js">var x = 5; $class.instanceOf(x, Number)</code>
 * will return true, but <code class="brush:js">var x = 5; x instanceof Number</code> will return false.
 * </p>
 * 
 * <p>
 * The $class library supplies dummy classes {@link Null} and {@link Undefined}. This method will identify
 * <em>null</em> and <em>undefined</em> as "instances" of those classes (respectively).
 * </p>
 *
 * <p>
 * If you do not need the special handling of primitive values, and the <em>type</em> is never going to be an
 * interface, then you can simply use the JavaScript <em>instanceof</em> operator for any classes created by 
 * {@link $class}.
 * </p>
 *
 * @param {mixed} obj any object or primitive.
 * @param {Function|$interface} type A constructor function of some class, or an interface.
 * @return {Boolean} true if the object is an instance of the specified class (or implementation of the specified
 * interface).
 */
$class.instanceOf = function(obj, type) {
  if (type instanceof $interface) {
    return $class.implementationOf(obj, type);
  }

  switch (typeof obj) {
    case "object":
      return (obj instanceof type) ||
             // special case for null
             (obj === null && type == Null);

    case "number":
      return (type == Number);

    case "string":
      return (type == String);

    case "boolean":
      return (type == Boolean);

    case "function":
    default:
      return (obj instanceof type);

    case "undefined":
      return (type == Undefined);
  }
};

//##############################################################################

/**
 * <p>
 * Gets the name of the class of the specified object. If the object is not an instance of a class that was created
 * by {@link $class} or adapted with {@link $class#adapt}, then it will be identified generically as "Object".
 * </p>
 *
 * <p>
 * JavaScript primitive values will be reported as being instances of their corresponding classes.
 * </p>
 * 
 * <p>
 * The $class library supplies dummy classes {@link Null} and {@link Undefined}. This method will identify
 * <em>null</em> and <em>undefined</em> as "Null" and "Undefined" (respectively).
 * </p>
 *
 * <p>
 * <h2>Example:</h2>
<pre class="brush:js">
$namespace("nested.namespace");

$class("nested.namespace.Example", {});

var example = new Example();

alert($class.typeOf(example)) // alerts "nested.namespace.Example"
alert($class.typeOf(5)) // alerts "5"
alert($class.typeOf(null)) // alerts "Null"
</pre>
 * </p>
 *
 * @param {mixed} obj any object or primitive.
 * @return {String} the name of the class of the object.
 */
$class.typeOf = function(obj) {
  return $class.getClass(obj).getName();
};

//##############################################################################

/**
 * <p>
 * Gets the {@link $class} object that represents the class of the specified object. If the object is not an 
 * instance of a class that was created by {@link $class} or adapted with {@link $class#adapt}, then Object's
 * $class object will be returned.
 * </p>
 *
 * <p>
 * For JavaScript primitive valuesthe $class objects of of their corresponding classes will be returned.
 * </p>
 *
 * <p>
 * For <em>null</em> and <em>undefined</em>, the $class objects for the dummy classes {@link Null} and 
 * {@link Undefined} will be returned (respectively).
 * </p>
 *
 * @param {mixed} obj any object or primitive.
 * @return {$class} The object's class's $class object.
 */
$class.getClass = function(obj) {
  if (obj == null) {
    if (obj === undefined) {
      return Undefined.$class;
    }

    return Null.$class;
  }

  return obj._$class || Object.$class;
};

//##############################################################################

/**
 * <p>
 * Creates a new instance of a specified class using an array of arguments.
 * </p>
 *
 * <p>
 * For "native" JavaScript classes (those defined by the ECMAScript spec, such as Number, RegExp, etc), the 
 * result of calling the constructor as a function will be returned. For classes that have corresponding primitive
 * types (Number, Boolean, String), this means that a primitive value will be returned instead of an instance of
 * the class. This is usually preferred for reasons such as instances of Strings, Boolean and Number always
 * evaluating to a boolean value of <em>true</em> even if their values are false-like.
 * </p>
 *
 * <p>
 * <h2>Example:</h2>
<pre class="brush:js">
$class("Person", {
  $constructor: function(firstName, lastName) {
    this._firstName = firstName;
    this._lastName) = lastName);
  },
  
  getFullName: function() {
    return this._firstName + " " + this._lastName;
  }
});

var person = $class.instantiate(Person, ["John", "Doe"]);
alert(person.getFullName()) // alerts "John Doe"

var number = $class.instantiate(Number, ["123"]);
alert(number) // alerts "123"
alert(typeof number) // alerts "number"
alert(number instanceof Number) // alerts "false"
</pre>
 * </p>
 *
 * @param {Function} ctor The constructor function for some class.
 * @param {Array} args An array of argument values to be passed to the constructor.
 * @return {mixed} A new instance of the specified class (or a primitive value).
 */
$class.instantiate = function(ctor, args) {
  if (ctor.$class && ctor.$class.isNative()) {
    return ctor.apply($class_library._globalObject, args);
  } else {
    $class_library._surrogateCtor.prototype = ctor.prototype;
    var result = new $class_library._surrogateCtor();
    ctor.apply(result, args);

    return result;
  }
};

//##############################################################################

/**
 * <p>
 * Creates an interface with the specified name and properties. Note that this documentation describes the behavior 
 * of $interface when used as a method (without the "new" keyword). This should only be called as a method rather 
 * than as a constructor. Instances of $interface are created as a private implementation detail of $interface being 
 * called as a method.
 * </p>
 * 
 * <p>
 * An interface describes methods that must be implemented by a class that claims to implement the interface.
 * Interfaces can also extend other interfaces and contain static properties/methods. 
 * See {@link $interface.descriptor} for details.
 * </p> 
 *
 * @class This class encapsulates information about an interface, but its constructor is used as a function 
 * (without the "new" keyword) to create an interface. It is used as a constructor for private implementation 
 * purposes only.
 *
 * @constructor
 *
 * @param {String} name The name of the interface. The name may describe a path of multiple nested 
 * namespace/class/object names separated by a period (.). The entire path to the interface name is expected 
 * to exist already. See {@link $namespace} for details about creating namespaces. 
 *
 * @param {$interface.descriptor} descriptor An object containing the properties of the interface, which can 
 * describe methods, static methods, static properties, and more.
 *
 * @return {$interface} The newly created interface.
 */
function $interface(name, descriptor) {
  // if being called as a constructor...
  if (this instanceof $interface) {
    if (!descriptor.calledFrom$interface) {
      $class_library._error("Error in attempt to define interface [" + name + "]: "
                          + "Do not directly instantiate $interface! Call $interface as a function instead.");
    }

    this._name = name;
    this._methods = {};
    this._methodsArray = null;
    this._interfaces = {};
    this._interfaces[name] = this;
    return;
  }

  var iface = new $interface(name, {calledFrom$interface: true});

  // extend interfaces
  if ($class_library._hasOwnProperty(descriptor, "$extends")) {
    var ifaces = descriptor.$extends;

    // convert to an array if it is a single object
    if (!(ifaces instanceof Array)) {
      ifaces = [ifaces];
    }

    for (var i = 0, ifacesLength = ifaces.length; i < ifacesLength; ++i) {
      // make sure the 'interface' to extend is really an interface
      if (!(ifaces[i] instanceof $interface)) {
        if (ifaces[i] == null) {
          $interface._propertyError(name, "$extends[" + i +"]", "Cannot extend null. You probably forgot to include the JS file containing the declaration of the parent class.");
        } else {
          $interface._propertyError(name, "$extends[" + i +"]", "$interface or array of $interfaces expected");
        }
      }

      iface._extend(ifaces[i]);
    }
  }

  var specialProperties = {$extends:true};
  var invalidProperties = {$constructor:true,$implements:true,$static:true};

  // process all properties in the descriptor
  for (var propertyName in descriptor) {
    // skip over the special properties
    if ($class_library._hasOwnProperty(specialProperties, propertyName)) {
      continue;
    }

    // check for invalid properties
    if ($class_library._hasOwnProperty(invalidProperties, propertyName)) {
      $interface._propertyError(name, propertyName, "invalid property for an $interface");
    }

    var value = descriptor[propertyName];

    if (value instanceof $class._ModifiedProperty && value.getModifier() == "static") {
      iface._addStatic(propertyName, value.getValue());
    } else if (value instanceof Function) {
      iface._addMethod(propertyName);
    } else {
      $interface._propertyError(name, propertyName, "Function or static value expected");
    }
  }

  // store the interface so it can be referenced by the specified name
  try {
    return eval(name + " = iface;");
  } catch (error) {
    $interface._error(name, "Invalid interface name: " + error.message);
  }
}

//##############################################################################

$interface.prototype = {
  //############################################################################

  /**
   * Gets the name of the interface.
   * @return {String} the full name of the interface.
   */
  getName: function() {
    return this._name;
  },

  //############################################################################

  /**
   * Tests if the interface (or any inherited interfaces) contains a method. 
   * @param {String} methodName the name of a potential method.
   * @return {Boolean} true if this interface has a method with the specified name.
   */
  hasMethod: function(methodName) {
    return Boolean(this._methods[methodName]);
  },

  //############################################################################

  /**
   * Gets a list of all methods in this interface (including methods of inherited interfaces).
   * @return {String[]} an array of method names.
   */
  getMethods: function() {
    if (!this._methodsArray) {
      this._methodsArray = [];

      for (var methodName in this._methods) {
        this._methodsArray.push(methodName);
      }
    }

    return this._methodsArray;
  },

  //############################################################################

  /**
   * <p>
   * Applies this interface to the specified object or class. This marks the object or class as implementing this
   * interface. The debug version of the $class library will verify that all interface methods are actually 
   * implemented by the object or class.
   * </p>
   *
   * <p>
   * This method has two intended uses:
   * <ol>
   *   <li>
   *     Retroactively mark a class as implementing an interface. This allows you to avoid modifying the definition
   *     of a class. You can pass either the constructor of the class, or the class's $class object 
   *     to applyInterface. Interfaces can be applied to classes that are not within the $class framework by 
   *     applying the interface to the class's constructor, but requires that all interface methods are implemented
   *     on the class's prototype rather than assigned directly to properties in the constructor. The interface
   *     information will be preserved if the class is later adapted into the $class framework via
   *     {@link $class.adapt}.
   *   </li>
   *   <li>
   *     Mark a specific object as implementing an interface. This could be an anonymous object, which would be 
   *     like creating an anonymous implementation of the interface. IMPORTANT: Interfaces cannot be applied to
   *     directly to functions. Doing so will be interpretted as an attempt to apply the interface to a class that
   *     is not within the $class framework (see above).
   *   </li>
   * </ol>
   * </p>
   *
   * <p>
   * When used to apply an interface to an object (not a class), be aware that a property (_$class_interfaces) will
   * be added to the object. If the object cannot be polluted with an extra property, then use {@link #wrapObject}.
   * </p>
   *
   * <p>
   * <h2>Debug version errors:</h2>
   * <ul>
   *   <li>Interface method not implemented in the class or object. The error message will list all methods not 
   *       implemented.</li>
   * </ul>
   * </p>
   *
   * @param {$class|Function|Object} obj The constructor or $class object of a class 
   * that implements all methods of this interface (to apply the interface to a class), or any object that 
   * implements all methods of this interface to apply the interface to that particular object.
   * @return {$class|Function|Object} The <em>obj</em> argument.
   */
  applyInterface: function(obj) {
    var classInfo = obj instanceof $class ? obj : obj.$class;

    if (classInfo) {
      obj = classInfo.getPrototype();
    } else if (obj.constructor == Function) {
      obj = obj.prototype;
    }

    var missingInterfaceMethods = new Array();

    for (var methodName in this._methods) {
      if (!(obj[methodName] instanceof Function)) {
        var ifaceName = this._methods[methodName];
        missingInterfaceMethods.push(ifaceName + "." + methodName);
      }
    }

    if (missingInterfaceMethods.length > 0) {
      var message = "Error in " + this.getName() + ".applyInterface(" + (classInfo ? classInfo.getName() : "[object]") + "); The following interface methods must be implemented: ";

      for (var i = 0; i < missingInterfaceMethods.length; ++i) {
        message += "[" + missingInterfaceMethods[i] + "] ";
      }

      $class_library._error(name, message);
    }

    if (classInfo) {
      this._applyInterface([classInfo]);
    } else {
      if (!obj._$class_interfaces) {
        obj._$class_interfaces = {};
      }

      for (var ifaceName in this._interfaces) {
        obj._$class_interfaces[ifaceName] = this._interfaces[ifaceName];
      }
    }

    return obj;
  },

  //############################################################################

  _applyInterface: function(classes) {
    for (var i = 0, classInfo; classInfo = classes[i]; ++i) {
      for (var ifaceName in this._interfaces) {
        classInfo._interfaces[ifaceName] = this._interfaces[ifaceName];
      }

      this._applyInterface(classInfo._children);
    }
  },

  //############################################################################

  /**
   * <p>
   * Creates an anonymous implementation of this interface whose methods are all implemented by simply calling the 
   * method of the name on the specified object. This may be an alternative to applying the interface directly to 
   * the object, which will add a property to the object that may be undesirable 9see {@link #applyInterface}. The 
   * debug version of the $class library will verify that the object implements all methods of this interface.
   * </p>
   *
   * @param {Object} obj An object that implements all methods of this interface.
   * @return {Object} a new object that implements this interface by calling methods on the <em>obj</em> argument. 
   */
  wrapObject: function(obj) {
    var result = {_obj: obj};

    result.toString = new Function("return '[" + $class.typeOf(obj) + " wrapped by " + this.getName() + "]';");

    for (var methodName in this._methods) {
      // Only create the interface method if a corresponding method exists on the object. This allows applyInterface to fail.
      if (typeof obj[methodName] == "function") {
        result[methodName] = new Function("return this._obj['" + methodName + "'].apply(this._obj, arguments);");
      }
    }

    return this.applyInterface(result);
  },

//############################################################################

  /**
   * Overridden to return a string that identifies this $interface.
   * @return {String} a string in the form "[$interface _name_]", where _name_ is the full name of this interface.
   */
  toString: function() {
    return "[$interface " + this._name + "]";
  },

  //############################################################################

  _extend: function(iface) {
    $class_library._copyObject(this._methods, iface._methods);
    $class_library._copyObject(this._interfaces, iface._interfaces);
  },

  //############################################################################

  _addStatic: function(propertyName, value) {
    this[propertyName] = value;
  },

  //############################################################################

  _addMethod: function(propertyName) {
    this._methods[propertyName] = this._name;
  }

  //############################################################################
};

//##############################################################################

$interface._error = function(name, message) {
  $class_library._error("Error in $interface(\"" + name + "\"" + ", ...): " + message);
};

//##############################################################################

$interface._propertyError = function(name, propertyName, message) {
  $class_library._error("Error in $interface(\"" + name + "\"" + ", ...), property [" + propertyName + "]: " + message);
};

//##############################################################################

/**
 * <p>
 * Creates an abstract method in a class. This is a property modifier for use within a 
 * {@link $class.descriptor}. 
 * </p>
 * <p>
 * An abstract method has no implementation and prevents the class from being instantiated. A class with
 * any abstract methods is inherently abstract; the class itself does not need to be (and cannot be) marked 
 * as absract. Any class that extends an abstract class is also abstract, unless it implements (overrides) all
 * inherited abstract methods. The abstract modifier may only be applied to non-{@link $final} and 
 * non-{@link $static} methods.
 * </p>
 *
 * <p>
 * <h2>Debug version errors:</h2>
 * <ul>
 *   <li>Attempt to instantiate an abstract class. The error message will indicate the abstract class and will 
         list all abstract methods and from which class they were inherited.</li>
 *   <li>Attempt to create an abstract property (not a method). The error message will indicate the class and 
 *       property name.</li>
 * </ul>
 * </p>
 *
 * <p>
 * <h2>Example:</h2>
<pre class="brush:js">
// abstract class that implements some algorithm, 
// but leaves some details unspecified
$class("AbstractBase", {
  // returns an array of numbers
  getData: $abstract(function() {}),
  
  // "protected" implementation detail of some algorithm.
  _shouldAddValue: $abstract(function(value) {}),
  
  // the algorithm that uses the abstract methods
  performSomeAlgorithmWithData: $final(function() {
    var data = this.getData();
    var result = 0;
    
    for (var i = 0; i < data.length; ++i) {
      if (this._shouldAddValue(data[i]) {
        result += data[i];
      }
    }
    
    return result;
  })
});

// class that implements the abstract methods
$class("Implementation", {
  $extends: AbstractBase,
  
  $constructor: function(data) {
    this._data = data;
  },
  
  getData: function() {
    return this._data;
  },
  
  _shouldAddValue: function(value) {
    return value > 0;
  }
});

var abstractBase = new AbstractBase(); // ERROR!

var impl = new Implementation([1, -5, 9, -10, 5]);
alert(impl.performSomeAlgorithmWithData()); // alerts "15"
</pre>
 * </p>
 *
 * @param {Function} method A function. The function itself is not actually used, but it must be provided.
 * This both allows for a sanity check that you aren not trying to create an abstract property (can't be done)
 * and provides a convenient place for you to declare/document the arguments.
 *
 * @return {Object} Something magical that, when assigned to a property of a {@link $class.descriptor}, causes
 * an abstract method to be created.
 */
function $abstract(method) {
  return new $class._ModifiedProperty("abstract", method);
}

/**
 * <p>
 * Creates a static property or method in a class or interface. This is a property modifier for use within a 
 * {@link $class.descriptor} or {@link $interface.descriptor}. 
 * </p>
 * <p>
 * A static property/method is accessible as a property of the class or interface without having an instance.
 * Within static methods, the "this" keyword will refer to the class or interface to which the static method 
 * belongs. Static methods/properties are not inherited; you must access them on the exact class or interface 
 * in which they were defined. Note that static methods by definition cannot be made {@link $abstract} or 
 * {@link $final}.
 * </p>
 *
 * <p>
 * <h2>Example:</h2>
<pre class="brush:js">
$class("Sample", {
  $constructor: function(name, type) {
    this._name = name;
    this._type = type;
  },
  
  equalTo: function(other) {
    return other 
        && other instanceof Sample
        && other._name == this._name
        && other._type == this._type;
  },
  
  // static constants for indicating the "type"
  Type: $static({
    GOOD: 1,
    BETTER: 2,
    BEST: 3
  }),
  
  // static factory method
  createBestSample: $static(function(name) {
    // "this" in static methods refers to the class/constructor
    return new this(name, this.Type.BEST);
  })
});

var sample1 = new Sample("Bob", Sample.Type.BEST);
var sample2 = Sample.createBestSample("Bob");

alert(sample1.equalTo(sample2)); // alerts "true"
</pre>
 * </p>
 *
 * @param {mixed} value Any value. May be a function.
 *
 * @return {Object} Something magical that, when assigned to a property of a {@link $class.descriptor} or 
 * {@link $interface.descriptor}, causes a static property to be created.
 */
function $static(value) {
  return new $class._ModifiedProperty("static", value);
}

/**
 * <p>
 * Creates a final method in a class, or creates a final class. This is a property modifier for use within a 
 * {@link $class.descriptor}. 
 * </p>
 * <p>
 * A final method is not allowed to be overridden by descendent classes. The final modifier can also be applied
 * to the {@link $class.descriptor#$constructor} of a class to prevent the class from being overridden 
 * althogether. The final modifier may only be applied to non-{@link $abstract} and non-{@link $static} methods.
 * </p>
 * <p>
 *
 * <h2>Debug version errors:</h2>
 * <ul>
 *   <li>Attempt to extend a class whose constructor is final.</li>
 *   <li>Attempt to override an inherited final method.</li>
 * </ul>
 * </p>
 *
 * <p>
 * <h2>Example:</h2>
<pre class="brush:js">
// extend this class to inherit a unique ID
$class("UniquelyIdentifiable", {
  $constructor: function() {
    this._id = "UniquelyIdentifiable_" + (UniquelyIdentifiable._idCounter++);
  },
  
  // If this method were ever overridden to do something different, it could
  // break the contract of this class (promises a unique ID). Making it final
  // keeps it completely within control of this class.
  getId: $final(function() {
    return this._id;
  }),
  
  _idCounter: $static(0)
});
</pre>
 * </p>
 *
 * @function
 *
 * @param {Function} method The implementation of the final method. 
 *
 * @return {Object} Something magical that, when assigned to a property of a {@link $class.descriptor), causes a final
 * method to be created.
 */
function $final(method) {
  return new $class._ModifiedProperty("final", method);
}

//##############################################################################

(function() {
  var nativeClassNames = [
    "Object",
    "Array",
    "String",
    "Number",
    "Boolean",
    "Function",
    "RegExp",
    "Date",

    "Error",
    "EvalError",
    "RangeError",
    "ReferenceError",
    "SyntaxError",
    "TypeError",
    "URIError"
  ];

  for (var i = 0, className; className = nativeClassNames[i]; ++i) {
    $class.adapt(this[className], className, true);
  }
})();

$class.adapt($namespace, "$namespace");
$class.adapt($class,     "$class");
$class.adapt($interface, "$interface");

//##############################################################################

/**
 * @name Undefined
 * @class
 * A dummy class used by various $class libray utility methods to represent the type of <em>undefined</em>
 * @see $class.typeOf
 * @see $class.instanceOf
 * @see $class.getClass
 */
$class("Undefined", {
  Undefined: $final(function() {
    $class_library._error("Attempted instantiation of the Undefined class.");
  })
});

/**
 * @name Null
 * @class
 * A dummy class used by various $class libray utility methods to represent the type of <em>null</em>
 * @see $class.typeOf
 * @see $class.instanceOf
 * @see $class.getClass
 */
$class("Null", {
  Null: $final(function() {
    $class_library._error("Attempted instantiation of the Null class.");
  })
});

//##############################################################################

/**
 * @name $class.descriptor
 * @class
 * <p>
 * This is a fake class used to document the descriptor object that is passed to the {@link $class} 
 * function to define a class. The documented properties are handled specially as described by each property.
 * All possible property names beginning with a dollar sign ($), or an underscore and dollar sign (_$) are to 
 * be considered reserved for use by the $class library. Anything else is fair game for defining properties 
 * and methods in a class.
 * </p>
 *
 * <p>
 * Unless documented otherwise, all properties in the descriptor are simply copied to the prototype of the 
 * new class, making them instance properties or instance methods. It is recommended that you avoid creating 
 * instance properties that are not methods. Instance properties of the class should generally be defined 
 * and initialized in the constructor to avoid non-obvious bugs that can arise from multiple instances 
 * unintentionally sharing a reference to an array or object through a prototype property.
 * </p>
 *
 * <p>
 * <h2>Property Modifiers</h2>
 * The following modifiers can be applied to properties in the descriptor. See the documentation for each 
 * modifier for details and examples.
 * <ul>
 *   <li>{@link $static} - create a static property or method.</li>
 *   <li>{@link $final} - prevent a method from being overridden (can also be applied to the {@link #$constructor}).</li>
 *   <li>{@link $abstract} - create an abstract method.</li>
 * </ul>
 * You may have noticed that there is no "private" modifier. I know that there's techniques for making 
 * "private fields" that truly cannot be accessed externally, but I don't believe the benefit outweighs the cost
 * in overhead and reduced maintainability of the code (it's hard to recognize which identifiers are actually 
 * private fields vs local or global variables). I prefer the simplicity of a leading underscore (_) for the names
 * of private fields and methods.
 * </p>
 *
 * <p>
 * <a name="$base"></a>
 * <h2>this.$base</h2>
 * In any non-static method (including the constructor), a special property named "$base" can be used to call
 * the overridden implementation of the same method. The overhead of supporting this special property 
 * is only added if the method contains a reference to this.$base (as determined by a simple regular expression). 
<pre class="brush:js">
$class("Person", {
  $constructor: function(name) {
    this._name = name;
  },
  
  getName: function() {
    return this._name;
  }
});

$class("Officer", {
  $extends: Person,
  
  $constructor: function(name, rank) {
    // call Person's constructor implementation
    this.$base(name);
    this._rank = rank;
  },
  
  getName: function() {
    // add the officer's rank to the name
    return this._rank + " " + this.$base();
  }
});

var officer = new Officer("Coltraine", "Sheriff");
alert(officer.getName()); // alerts "Sheriff Coltraine"
</pre>
 * </p>
 */
$class_library._$class_descriptor_docs = 
/**
 * @lends $class.descriptor.prototype
 */
{
  /**
   * <p>
   * A function that will be used to implement the constructor of the class. A constructor is not required. If
   * not specified, a default constructor will be created. Default constructors will automatically call the 
   * parent class's constructor with all the arguments that were passed in. Use the special 
   * <a href="#$base">this.$base</a> property to 
   * call the parent class's constructor (a.k.a., initializing the parent class) with specific arguments. If a 
   * constructor is provided, but there is no usage of this.$base, the parent class's constructor will be 
   * automatically called before your constructor implementation (with all the same arguments).
   * </p>
   *
   * <p>
   * The constructor may be modified with the {@link $final} modifier to make the entire class final (prevents
   * other classes from extending this class).
   * </p>   
   *
   * <p>
   * The constructor can also be provided as a property whose name is the same as the class's name.
   * </p>   
   *
   * <p>
   * <h2>Debug version warnings:</h2>
   * <ul>
   *   <li>No constructor specified; using default constructor.</li>
   * </ul>
   * </p>
   *
   * <p>
   * <h2>Debug version errors:</h2>
   * <ul>
   *   <li>Constructor value is not a function.</li>
   * </ul>
   * </p>
   *
   * <p>
   * <h2>Optimized trivial constructors:</h2>
   * If there nothing in the body of the constructor function (no comments; only whitespace), or if no 
   * constructor is provided, then it is considered to be a trivial constructor. In the non-debug version of 
   * the $class library, trivial constructors are skipped when instantiating classes with a chain of 
   * inheritance. For example, let's say there is an inheritance chain of classes: A <- B <- C <- D <- E 
   * (A is the base class in this chain). If A, B and D all have trivial constructors, then the instantiation 
   * of B will simply execute a single empty function. Instantiating C will execute only C's constructor.
   * Instantiating D will execute a wrapper function that simply call's C's constructor. Instantiating E will
   * execute E's constructor, then skip right to C's constructor (even if E's constructor explicitly calls 
   * <a href="#$base">this.$base</a>).
   * </p>
   *
   * <p>
   * <h2>Example:</h2>
<pre class="brush:js">
$namespace("Shapes");

$class("Shapes.Rectangle", {
  // this is a constructor
  $constructor: function(width, height) {
    this._width = width;
    this._height = height;
  }
});

$class("Shapes.Square", {
  $extends: Rectangle,

  // this is a constructor too
  Square: function(size) {
    // Call Square's constructor implementation
    this.$base(size, size);
  }
});
</pre>
   * </p>
   *
   * @type Function
   */
  $constructor: null,
  
  /**
   * <p>
   * A class to be inherited by this class. Instances of this class will be considered to be instances of the 
   * specified parent class by both the JavaScript "instanceof" keyword, and {@link $class#instanceOf}. All
   * instance methods and properties will be inherited. This class will be considered to implement all 
   * {@link $interface}s that are implemented by the parent class. 
   * </p>
   *
   * <p>
   * See {@link #$constructor} for details about initializing the parent class.
   * </p>
   *
   * <p>
   * <h2>Debug version errors:</h2>
   * <ul>
   *   <li>$extends is specified, but the value is null/undefined. This usually indicates that you attempted
   *       to extend a class and forgot to include the JS file that defines that class, or maybe you misspelled
   *       the class name.</li>
   *   <li>Attempt to extend something that is not a Function.</li>
   *   <li>Attempt to extend a class whose {@link #$constructor} is {@link $final}.</li>
   *   <li>Attempt of a non-multiton class to extend a multiton ({@link $class.descriptor#$multiton}).</li>
   *   <li>Attempt of a non-singleton class to extend a singleton ({@link $class.descriptor#$singleton}).</li>
   * </ul>
   * </p>
   *
   * <p>
   * <h2>Debug version warnings:</h2>
   * <ul>
   *   <li>The class being extended is being automatically adapted into the $class framework with a generated
   *       name (see {@link $class.adapt})</li>
   * </ul>
   * </p>
   *
   * <p>
   * <h2>Example:</h2>
<pre class="brush:js">
$class("Rectangle", {
  // this is a constructor
  $constructor: function(width, height) {
    this._width = width;
    this._height = height;
  },
  
  getArea: function() {
    return this._width * this.height;
  }
});

$class("Square", {
  $extends: Rectangle,

  // this is a constructor too
  Square: function(size) {
    // Call Square's constructor implementation
    this.$base(size, size);
  }
});

var square = new Square(10);

alert(square instanceof Square); // alerts "true"
alert(square instanceof Rectangle); // alerts "true"
alert(square.getArea()); // alerts "100"
</pre>
   * </p>
   *
   * @type Function
   *
   * @see $class#instanceOf
   */
  $extends: null,
  
  /**
   * <p>
   * One or more {@link $interface}s that are implemented by this class. All methods of all interfaces must be
   * present in this class. Inherited methods and {@link $abstract} methods both satisfy this requirement.
   * </p>
   *
   * <p>
   * <h2>Debug version errors:</h2>
   * <ul>
   *   <li>Interface method not implemented in the class. The error message will list all methods not 
   *       implemented.</li>
   * </ul>
   * </p>
   *
   * <p>
   * <h2>Example:</h2>
<pre class="brush:js">
$interface("Identifiable", {
  getId: function() {}
});

$interface("Named", {
  getName: function() {}
});

$class("Person", {
  $implements: [Identifiable, Named],
  
  $constructor: function(governmentId, name) {
    this._governmentId = governmentId;
    this._name = name;
  },
  
  getId: function() {
    return this._governemtnId;
  },
  
  getName: function() {
    return this._name;
  }
});
</pre>
   * </p>
   *
   * @type ($interface|$interface[])
   *
   * @see $class#implementsInterface
   * @see $class#implementionOf
   * @see $class#instanceOf
   */
  $implements: null,
  
  /**
   * <p>
   * A static initializer function that is executed immediately after the class is created. This can be used 
   * to perform complex initialization of {@link $static} properties. Like static methods, the static
   * initializer is executed within the context of the class/constructor.
   * </p>
   *
   * <p>
   * <h2>Example:</h2>
<pre class="brush:js">
$class("Example", {
  ValuesMap: $static({
    ZERO: 0,
    ONE: 1,
    TWO: 2,
    THREE: 3
  }),
  
  // initialized in the static initializer
  ValuesReverseMap: $static({}),
  
  $static: function() {
    for (var i in this.ValuesMap) {
      this.ValuesReverseMap[this.ValuesMap[i]] = i;
    }
  }
});

// static initializer has already been executed by this point.
// alerts "true"
alert(Example.ValuesReverseMap[Example.ValuesMap["TWO"]] == "TWO");
</pre>
   * </p>
   *
   * @type Function
   */
  $static: null,
  
  /**
   * <p>
   * By default, every class created by {@link $class} will have its own toString implementation that returns a 
   * string in the form "[object _name_]", where _name_ is the full name of the class. This can be useful for 
   * debugging. toString can be overridden with a custom implementation in any class.
   * </p>
   *
   * @type Function
   */
  toString: null,
  
  /**
   * <p>
   * Configures a class as a singleton. A singleton cannot be directly instantiated. A static "getInstance" method
   * is automatically created that will take care of instantiating the class the first time and returning the same
   * instance from all subsequent calls.
   * </p>
   *
   * <p>
   * The value of this property can either be an anonymous {@link $class.descriptor.singleton_config} object, or 
   * simply a boolean "true" if the defaults of the config object are acceptable.
   * </p>
   *
   * <p>
   * The advantages of creating a singleton this way (as opposed to simply creating a global named object with some
   * methods and properties) is that it defers initialization until the first use in code, and it can extend another
   * class, implement interfaces, etc.
   * </p>
   *
   * <p>
   * A singleton can extend any non-{@link $final} class that is not a multiton ({@link $class.descriptor#$multiton}). 
   * Yes, a singleton can extend another singleton. Instances are not shared between parent and child singleton, so by 
   * extending singleton, you will technically end up with two equivalent instances of the parent class (one of 
   * them actually being an instance of the child class). Non-singleton classes cannot extend a singleton.
   * What does it mean for a singleton to extend another singleton? I dunno; you tell me. I couldn't absolutely 
   * convince myself to disallow it, so here it is.
   * </p>
   *
   * <p>
   * <h2>Warning:</h2>
   * If you currently only need one instance and you're just being lazy and designing your class as a singleton so
   * that you can call getInstance() in multiple places instead of passing a reference around, you should really 
   * just spend the little bit of extra time to instantiate your class in one place and pass it to the constructors
   * of the various classes that need to reference it. You'll thank me some day when you need to reuse your code 
   * some day in a way such that your singleton needs to be dynamically reconfigured or you actually find that you
   * really do need multiple simultaneous instances with different configurations. It's much easier, cleaner and 
   * less error-prone to reinstantiate objects with new data than it is to hack up a singleton to reconfigure 
   * itself.
   * </p>
   *
   * <p>
   * <h2>Example:</h2>
<pre class="brush:js">
$class("Example", {
  $singleton: true,
  
  $constructor: function() {
    this._value = 0;
  },
  
  getValue: function() {
    return this._value;
  },
  
  incrementValue: function() {
    ++this._value;
  }
});

var example1 = Example.getInstance();
var example2 = Example.getInstance();

alert(example1 == example2); // alerts "true"
alert(example1.getValue()); // alerts "0"
alert(example2.getValue()); // alerts "0"

example1.incrementValue();
alert(example2.getValue()); // alerts "1"
</pre>
   * </p>
   *
   * @type ($class.descriptor.singleton_config|Boolean)
   */
  $singleton: null,
  
  /**
   * <p>
   * Configures a class as a multiton. A multiton cannot be directly instantiated. A static "getInstance" method
   * is automatically created that will take care of instantiating the class the first time with a given set of and
   * arguments and returning the same instance from all subsequent calls with the same set of arguments.
   * </p>
   *
   * <p>
   * The value of this property can either be an anonymous {@link $class.descriptor.multiton_config} object, or 
   * simply a boolean "true" if the defaults of the config object are acceptable.
   * </p>
   *
   * <p>
   * Multitons are generally good for classes that are expensive to instantiate and if only a relatively small 
   * number of instances will each be referenced multiple times, but you want to avoid the hassle of maintaining 
   * references to instances. This is a very simple caching system that can provide performance improvements, but 
   * does not allow for any control over the cache, such as enforcing a limit to the cache.
   * </p>
   *
   * <p>
   * A multiton can extend any non-{@link $final} class that is not a singleton ({@link $class.descriptor#$singleton}).
   * Yes, a multiton can extend another multiton. Instances are not shared between parent and child multitons, so by 
   * extending multitons, you can technically end up with two equivalent instances of the parent class (one of
   * them actually being an instance of the child class). Non-multiton classes cannot extend a multiton.
   * What does it mean for a multiton to extend another multiton? I dunno; you tell me. I couldn't absolutely 
   * convince myself to disallow it, so here it is.
   * </p>
   *
   * <p>
   * <h2>Warning:</h2>
   * By default, the cache key is built very naively. If there is any way that different parameter values could
   * produce identical instances of your class, then you must provide your own cache key algorithm. This problem
   * is illustrated in the example. See {@link $class.descriptor.multiton_config#createCacheKey} for details about
   * providing your own cache key algorithm.
   * </p>
   * 
   * <p>
   * <h2>Example:</h2>
<pre class="brush:js">
$class("Example", {
  $multiton: true,
  
  // name defaults to "John Doe" if empty or null
  $constructor: function(name) {
    this._name = name || "John Doe";
  },
  
  getName() {
    return this._name;
  }
});

var example1 = Example.getInstance("John Doe");
var example2 = Example.getInstance("John Doe");

// same instance if instantiated with the same arguments
alert(example1 == example2) // alerts "true"

// instantiate with different arguments...
var example3 = Example.getInstance(null);

// different instance
alert(example2 == example3) // alerts "false"
// ... but the same name
alert(example2.getName() == example3.getName()) // alerts "false"
</pre>
   * </p>
   *
   * @type ($class.descriptor.multiton_config|Boolean)
   */
  $multiton: null
};

/**
 * @name $class.descriptor.singleton_config
 * @class
 * <p>
 * This is a fake class to document the configuration object for use with {@link $class.descriptor#$singleton}
 * to setup a class as a singleton. If the defaults are acceptable, you can simply provide a value of 
 * <em>true</em> instead of a singleton config object.
 * </p>
 */
$class_library._$singleton_config_docs = 
/**
 * @lends $class.descriptor.singleton_config.prototype
 */
{
  /**
   * <p>
   * A function that is used to create the singleton instance. This function receives no arguments and is 
   * expected to return an instance of the class. The function is called within the context of the 
   * class/constructor. Defaults to a function that simply instantiates the class with no constructor arguments.
   * </p>
   *
   * <p>
   * This property is useful for when you want to have a singleton, but it needs be instantiated with 
   * "dynamic static data"; data that is determined by the server so that it cannot be hard-coded in the JavaScript, 
   * but the data never changes during the lifetime of the script, so it is static as far as the script is 
   * concerned.
   * </p>
   *
   * <p>
   * Defaults to a function that simply instantiates the class with the same arguments.
   * </p>
   *
   * <p>
   * <h2>Warning:</h2>
   * If you think you want to do something like this, chances are that you really shouldn't be creating a singleton.
   * This pattern of initializing a singleton with "dynamic static data" should only be used if the purpose of the 
   * class inherently makes the design require that there never be more than one instance, because multiple 
   * instances would catastrophically interfere with each other or other data. See also the warning about 
   * {@link $class.descriptor#$singleton}.
   * </p>
   *
   * <p>
   * <h2>Example:</h2>
<pre class="brush:js">
$class("Example", {
  $singleton: {
    createInstance: function() {
      // A default value allows the code to not depend on an externally 
      // defined value 
      var paramValue = this.DEFAULT_PARAM_VALUE;
      
      // If a global variable "EXAMPLE_PARAM_VALUE" exists, use its value.
      // The value of this variable could have been rendered by server 
      // code (PHP, JSP, etc). This allows default behavior/configuration
      // to be overridden if desired.
      if (typeof EXAMPLE_PARAM_VALUE != "undefined") {
        paramValue = EXAMPLE_PARAM_VALUE;
      }      
      
      return new this(paramValue);
    },
  },
  
  $constructor: function(param) {
    this._param = param;
  },
  
  getParam: function() {
    return this._param;
  },
  
  DEFAULT_PARAM_VALUE: $static(10)
});

var EXAMPLE_PARAM_VALUE = 5;
alert(Example.getInstance().getParam()); // alerts "5"

EXAMPLE_PARAM_VALUE = 20;
alert(Example.getInstance().getParam()); // still alerts "5"
</pre>
   * </p>
   *
   * @type Function
   */
  createInstance: null
};

/**
 * @name $class.descriptor.multiton_config
 * @class
 * <p>
 * This is a fake class to document the configuration object for use with {@link $class.descriptor#$multiton}
 * to setup a class as a multiton. If the defaults are acceptable, you can simply provide a value of 
 * <em>true</em> instead of a multiton config object.
 * </p>
 */
$class_library._$multiton_config_docs = 
/**
 * @lends $class.descriptor.multiton_config.prototype
 */
{
  /**
   * <p>
   * A function that is used to create each multiton instance. This function receives all arguments that are 
   * passed to the class's static getInstance method and is expected to return an instance of the class. The 
   * function is called within the context of the class/constructor. Defaults to a function that simply 
   * instantiates the class with the same arguments that were passed to getInstance().
   * </p>
   *
   * <p>
   * I'm having a hard time coming up with a legitemate use for this property, but it seemed like it should be 
   * supported to maintain consistency with {@link $class.descriptor.singleton_config}. Good luck :)
   * </p>
   *
   * @type Function
   */
  createInstance: null,
  
  /**
   * <p>
   * A function that is used to create a cache key for looking up cached multiton instances. This function 
   * receives all arguments that are passed to the class's static getInstance method and is expected to return
   * a String that uniquely identifies the combination of arguments. The arguments should be "normalized", such 
   * that if two different sets of arguments would be interpretted the same and result in equivalent instances,
   * an identical cache key should be generated for those two different sets of arguments.  
   * </p>
   *
   * <p>
   * Defaults to a function that simply converts all arguments to Strings and joins them together with pipes (|).
   * </p>
   *
   * <p>
   * <h2>Example:</h2>
<pre class="brush:js">
$class("Example", {
  $multiton: {
    createCacheKey: function(name) {
      // the cache key must account for any kind of 
      // defaulting/processing of params in the constructor
      return name || "John Doe";
    }
  },
  
  // name defaults to "John Doe" if empty or null
  $constructor: function(name) {
    this._name = name || "John Doe";
  },
  
  getName() {
    return this._name;
  }
});

var example1 = Example.getInstance(null);
var example2 = Example.getInstance("John Doe");

// same name
alert(example1.getName() == example2.getName()) // alerts "true"
// same instance
alert(example1 == example2) // alerts "true"
</pre>
   * </p>
   *
   * @type Function
   */
  createCacheKey: null
};

/**
 * @name $interface.descriptor
 * @class
 * <p>
 * This is a fake class used to document the descriptor object that is passed to the {@link $interface} 
 * function to define an interface. The documented properties are handled specially as described by each property.
 * All possible property names beginning with a dollar sign ($), or an underscore and dollar sign (_$) are to 
 * be considered reserved for use by the $class library. Anything else is fair game for defining properties 
 * and methods in an interface.
 * </p>
 *
 * <p>
 * Unless documented otherwise, all properties in the descriptor are expected to describe interface methods. The 
 * value of these properties must be a function, but its implementation will not be used, so don't bother writing
 * code in the body of the function. Providing a function value for these properties gives you a good place to
 * define and document params and return values. It also allows the debug version of the $class library to yell at
 * you if you try to create a non-static non-method property on an interface.
 * </p>
 *
 * <p>
 * <h2>Property Modifiers</h2>
 * The following modifiers can be applied to properties in the descriptor. See the documentation for each 
 * modifier for details and examples.
 * <ul>
 *   <li>{@link $static} - create a static property or method.</li>
 * </ul>
 * </p>
 */
$class_library._$interface_descriptor_docs = 
/**
 * @lends $interface.descriptor.prototype
 */
{
  /**
   * <p>
   * One or more interfaces to be inherited by this interface. Implementations of this interface will be considered
   * to be implementations of all inherited interfaces. All interface methods will be inherited, meaning that  
   * implementations of this interface must implement all methods of all inherited interfaces in addition to any
   * methods defined by this interface.
   * </p>
   *
   * <p>
   * <h2>Debug version errors:</h2>
   * <ul>
   *   <li>$extends is specified, but one of the values is null/undefined. This usually indicates that you attempted
   *       to extend an interface and forgot to include the JS file that defines that interface, or maybe you 
   *       misspelled the interface name.</li>
   *   <li>Attempt to extend something that is not an $interface.</li>
   * </ul>
   * </p>
   *
   * @type $interface|$interface[]
   */
  $extends: null
};
