//##############################################################################
// The $class Library
// Copyright 2006, Jeff Lau
// License: http://creativecommons.org/licenses/LGPL/2.1/
// Contact: jlau@uselesspickles.com
// Web:     www.uselesspickles.com
//##############################################################################

/**
 * @fileOverview
 * <p>
 * This is the non-debug version of the $class library. It is slimmed down to the bare minimum to make all the
 * features work, but with no error checking. This signifigantly reduces overhead to keep your code running fast.
 * be sure to switch over to including class-debug.js when you are debugging or developing to make use of all
 * the error detection.
 * </p>
 */

$class_library = {
  version: "2.0b",
  
  debug: false,

  _globalObject: this,

  _hasOwnProperty: Object.prototype.hasOwnProperty ?
    function (obj, prop) {
      return obj.hasOwnProperty(prop);
    }
    :
    function (obj, prop) {
      return (prop in obj) && obj.constructor.prototype[prop] !== obj[prop];
    },

  _surrogateCtor: function() {
  }
};

//##############################################################################

function $namespace(name) {
  // if being called as a constructor...
  if (this instanceof $namespace) {
    this.toString = function() {
       return "[$namespace " + name + "]";
    }

    $class_library._surrogateCtor.prototype = this;
    return new $class_library._surrogateCtor();
  }

  var components = name.split(".");
  var context = $class_library._globalObject;

  for (var i = 0; i < components.length; ++i) {
    var nextContext = context[components[i]];

    if (!nextContext) {
      nextContext = new $namespace(components.slice(0, i + 1).join("."));
      context[components[i]] = nextContext;
    }

    context = nextContext;
  }
  
  return context;
}

//##############################################################################

function $class(name, descriptor) {
  // if being called as a constructor...
  if (this instanceof $class) {
    this._name = name;
    this._isNative = descriptor.isNative;
    this._ctor = descriptor.ctor;
    this._baseCtor = descriptor.baseCtor || (this._ctor == Object ? null : Object);
    this._interfaces = descriptor.interfaces || {};
    this._children = [];

    // if not creating $class for Object...
    if (this._ctor != Object) {
      // if inheriting something other than Object...
      if (this._baseCtor != Object) {
        // if this $class object is being created internally by the $class function...
        if (descriptor.calledFrom$class) {
          // a 'surrogate' constructor is used to create inheritance relationship
          // without actually invoking the base class's constructor code
          $class_library._surrogateCtor.prototype = this._baseCtor.prototype;
          this._ctor.prototype = new $class_library._surrogateCtor();
          this._ctor.prototype.constructor = this._ctor;
        }

        var base$class = this._baseCtor.$class;
        // inherit info about the base class
        // inlined object copy
        var copyTo = this._interfaces;
        var copyFrom = base$class._interfaces;
        for (var prop in copyFrom) { copyTo[prop] = copyFrom[prop]; }
      
        base$class._children.push(this);
      }
      
      // store this class info on the prototype
      this._ctor.prototype._$class = this;
    }
    // Storing class info on prototype and constructor causes an IE leak; store created $class objects to clean up when
    // unloading the page.    
    $class._cleanupCache.push(this);    

    return;
  }

  // local aliases to reduce scope and property lookups
  var wrapExtendedMethod = $class._wrapExtendedMethod;
  var objectProto = Object.prototype;

  var baseCtor = descriptor.$extends || Object;
  var namePath = name.split(".");
  var ctorName = namePath.pop();
  var ctor = descriptor.$constructor || descriptor[ctorName];

  if (!baseCtor.$class) {
    $class.adapt(baseCtor, name + "$base");
  }

  if (!ctor || $class._isTrivialCtor.test(ctor)) {
    if (baseCtor._$class_isTrivialCtor) {
      ctor = $class._createDefaultCtor();
      ctor._$class_isTrivialCtor = true; 
    } else {
      var relevantBase = baseCtor._$class_relevantImplementation || baseCtor;
      ctor = $class._createDefaultExtendedCtor(relevantBase);
      ctor._$class_relevantImplementation = relevantBase;
    }
  } else if (ctor.toString().indexOf("this.$base") != -1) {
    ctor = wrapExtendedMethod(ctor, baseCtor._$class_relevantImplementation || baseCtor);
  } else if (!baseCtor._$class_isTrivialCtor) {
    ctor = $class._wrapExtendedCtor(ctor, baseCtor._$class_relevantImplementation || baseCtor);
  }
  
  ctor.$class = new $class(name, {ctor:ctor, baseCtor:baseCtor, calledFrom$class:true});
  var proto = ctor.prototype;

  if (descriptor.$singleton) {
    var createInstance = (typeof descriptor.$singleton.createInstance == "function") ? descriptor.$singleton.createInstance : $class._createSingletonInstance_implementation;

    descriptor._$class_createSingletonInstance = $static(createInstance);
    descriptor._$class_singletonInstance = $static(null);
    descriptor.getInstance = $static($class._getSingletonInstance_implementation);
  }

  if (descriptor.$multiton) {
    var createInstance = (typeof descriptor.$multiton.createInstance == "function") ? descriptor.$multiton.createInstance : $class._createMultitonInstance_implementation;
    var createCacheKey = (typeof descriptor.$multiton.createCacheKey == "function") ? descriptor.$multiton.createCacheKey : $class._createMultitonCacheKey_implementation;

    descriptor._$class_createMultitonInstance = $static(createInstance);
    descriptor._$class_createMultitonCacheKey = $static(createCacheKey);
    descriptor._$class_multitonInstances = $static({});
    descriptor.getInstance = $static($class._getMultitonInstance_implementation);
  }

  // implement interfaces
  if (descriptor.$implements != null) {
    var ifaces = descriptor.$implements;

    // convert to an array if it is a single object
    if (!(ifaces instanceof Array)) {
      ifaces = [ifaces];
    }

    for (var i = 0, iface; iface = ifaces[i]; ++i) {
      // inlined object copy
      var copyTo = ctor.$class._interfaces;
      var copyFrom = iface._interfaces;
      for (var prop in copyFrom) { copyTo[prop] = copyFrom[prop]; }
    }
  }

  var specialProperties = $class._specialProperties;
  var processedProperties = {};
  specialProperties[ctorName] = true;

  // process all properties in the class descriptor
  for (var propertyName in descriptor) {
    // skip over special properties
    if (specialProperties[propertyName] === true) {
      continue;
    }

    var value = descriptor[propertyName];
    
    if (value != null && value._$class_modifiedProperty) {
      // In the release version, only the static modifier results in a
      // $class._ModifiedProperty object, so there is no need to check
      // the modifier value
      ctor[propertyName] = value.value;
    } else if (typeof value == "function" && value.toString().indexOf("this.$base") != -1) {
      // wrap the method to give it access to the special $base property
      proto[propertyName] = wrapExtendedMethod(value, proto[propertyName]);
    } else {
      proto[propertyName] = value;
    }
    
    processedProperties[propertyName] = true;
  }
  
  specialProperties[ctorName] = false;
  
  // These are property names that IE mistakenly skips while iterating over an object's properties (only the standard 
  // properties on Object's prototype are supposed to be skipped; a property of the same name that is created directly 
  // on an object should be iterated)
  var skippedByIE = $class._propertiesSkippedByIE;
  
  for (var i = 0, propertyName; propertyName = skippedByIE[i]; ++i) {
    var value = descriptor[propertyName];

    if (value !== objectProto[propertyName] && processedProperties[propertyName] !== true) {
      if (value.toString().indexOf("this.$base") != -1) {
        // wrap the method to give it access to the special $base property
        proto[propertyName] = wrapExtendedMethod(value, proto[propertyName]);
      } else {
        proto[propertyName] = value;
      }
    }
  }

  // store the constructor so it is accessible by the specified name
  var scope = $class_library._globalObject;
  
  for (var i = 0, pathComponent; pathComponent = namePath[i]; ++i) {
    scope = scope[pathComponent];
  } 
  
  var result = scope[ctorName] = ctor;

  //  call the static initializer
  if (descriptor.$static) {
    descriptor.$static.call(ctor);
  }

  return result;
}

//##############################################################################

$class.prototype = {
  getName: function() {
    return this._name;
  },

  isNative: function() {
    return this._isNative;
  },

  getPrototype: function() {
    return this._ctor.prototype;
  },

  getConstructor: function() {
    return this._ctor;
  },

  getSuperclass: function() {
    return this._baseCtor ? this._baseCtor.$class : null;
  },

  isInstance: function(obj) {
    return $class.instanceOf(obj, this._ctor);
  },

  implementsInterface: function(iface) {
    return $class_library._hasOwnProperty(this._interfaces, iface.getName());
  },

  toString: function() {
    return "[$class " + this._name + "]";
  }
};

//##############################################################################
// List of $class objects created so they can be cleaned up when unloading the page
$class._cleanupCache = [];

// using attachEvent because this is only needed to avoid memory leaks in IE, and only IE supports attachEvent.
if (typeof window != "undefined" && window.attachEvent) {
  window.attachEvent('onunload', function() {
    for (var i=0,clz; clz = $class._cleanupCache[i]; i++) { 
      delete clz._ctor.prototype._$class;
      delete clz._ctor.$class;
    }
  });
}

//##############################################################################

$class._isTrivialCtor = /^\s*function[^(]*\([^)]*\)[^{]*\{\s*(return)?\s*;?\s*\}\s*$/;

$class._specialProperties = {$constructor:true,$singleton:true,$multiton:true,$extends:true,$implements:true,$static:true};

$class._propertiesSkippedByIE = ["valueOf", "toString"];

//##############################################################################

$class._wrapExtendedMethod = function(method, baseMethod) {
  var result = function() {
    var h=this.$base;
    this.$base=baseMethod;
    try{return method.apply(this,arguments);}
    finally{this.$base=h;}
  };

  result.toString = function() { return method.toString(); };

  return result;
};

//##############################################################################

$class._createDefaultCtor = function() {
  return function() {};
};

//##############################################################################

$class._createDefaultExtendedCtor = function(baseMethod) {
  return function() {
    baseMethod.apply(this, arguments);
  };
};

//##############################################################################

$class._wrapExtendedCtor = function(method, baseMethod) {
  var result = function() {
    baseMethod.apply(this, arguments);
    method.apply(this, arguments);
  };

  result.toString = function() { return method.toString(); };

  return result;
};

//##############################################################################

$class._getSingletonInstance_implementation = function() {
  return this._$class_singletonInstance || (this._$class_singletonInstance = this._$class_createSingletonInstance());
};

//##############################################################################

$class._createSingletonInstance_implementation = function() {
  return new this();
};

//##############################################################################

$class._getMultitonInstance_implementation = function() {
  var cacheKey = this._$class_createMultitonCacheKey.apply(this, arguments);

  return this._$class_multitonInstances[cacheKey] || (this._$class_multitonInstances[cacheKey] = this._$class_createMultitonInstance.apply(this, arguments));
};

//##############################################################################

$class._createMultitonInstance_implementation = function() {
  return $class.instantiate(this, arguments);
};

//##############################################################################

$class._createMultitonCacheKey_implementation = function() {
  return Array.prototype.join.call(arguments, "|");
};

//##############################################################################

$class.adapt = (function() {
  var unnamedCount = 0;
  
  return function(ctor, name, isNative) {
    if (ctor.$class) {
      return false;
    }

    var baseCtor = ctor.prototype.constructor;

    if (ctor != Object) {
      var hold = ctor.prototype.constructor;

      if (delete ctor.prototype.constructor) {
        baseCtor = ctor.prototype.constructor;
        ctor.prototype.constructor = hold;
      }
    }

    if (!name) {
      name = "$unnamed_adapted_class_" + (unnamedCount++);
    }

    if (baseCtor == ctor || !baseCtor) {
      baseCtor = Object;
    } else if (!baseCtor.$class) {
      $class.adapt(baseCtor, name + "$base");
    }

    if ($class._isTrivialCtor.test(ctor)) {
      if (baseCtor._$class_isTrivialCtor) {
        ctor._$class_isTrivialCtor = true; 
      } else {
        var relevantBase = baseCtor._$class_relevantImplementation || baseCtor;
        ctor._$class_relevantImplementation = relevantBase;
      }
    }
    
    var interfaces = ctor.prototype._$class_interfaces;
    delete ctor.prototype._$class_interfaces;
    
    ctor.$class = new $class(name, {
      ctor: ctor, 
      baseCtor: baseCtor, 
      isNative: isNative, 
      interfaces: interfaces,
      calledFrom$class_adapt: true
    }); 

    return true;
  };
})();

//##############################################################################

$class.resolve = function(name) {
  var components = name.split(".");
  var context = $class_library._globalObject;

  for (var i = 0; i < components.length; ++i) {
    if (context == null) {
      return undefined;
    }
    
    var context = context[components[i]];
  }
  
  return context;
};

//##############################################################################

$class.implementationOf = function(obj, iface) {
  if (obj._$class_interfaces && $class_library._hasOwnProperty(obj._$class_interfaces, iface.getName())) {
    return true;
  }

  return $class.getClass(obj).implementsInterface(iface);
};

//##############################################################################

$class.instanceOf = function(obj, type) {
  if (type instanceof $interface) {
    return $class.implementationOf(obj, type);
  }

  switch (typeof obj) {
    case "object":
      return (obj instanceof type) ||
             // special case for null
             (obj === null && type == Null);

    case "number":
      return (type == Number);

    case "string":
      return (type == String);

    case "boolean":
      return (type == Boolean);

    case "function":
    default:
      return (obj instanceof type);

    case "undefined":
      return (type == Undefined);
  }
};

//##############################################################################

$class.typeOf = function(obj) {
  return $class.getClass(obj).getName();
};

//##############################################################################

$class.getClass = function(obj) {
  if (obj == null) {
    if (obj === undefined) {
      return Undefined.$class;
    }

    return Null.$class;
  }

  return obj._$class || Object.$class;
};

//##############################################################################

$class.instantiate = function(ctor, args) {
  if (ctor.$class && ctor.$class.isNative()) {
    return ctor.apply($class_library._globalObject, args);
  } else {
    $class_library._surrogateCtor.prototype = ctor.prototype;
    var result = new $class_library._surrogateCtor();
    ctor.apply(result, args);

    return result;
  }
};

//##############################################################################

function $interface(name, descriptor) {
  // if being called as a constructor...
  if (this instanceof $interface) {
    this._name = name;
    this._methods = {};
    this._methodsArray = null;
    this._interfaces = {};
    this._interfaces[name] = this;
    return;
  }

  var iface = new $interface(name);

  // extend interfaces
  if (descriptor.$extends != null) {
    var ifaces = descriptor.$extends;

    // convert to an array if it is a single object
    if (!(ifaces instanceof Array)) {
      ifaces = [ifaces];
    }

    for (var i = 0, ifacesLength = ifaces.length; i < ifacesLength; ++i) {
      // inlined object copy
      var copyTo = iface._methods; 
      var copyFrom = ifaces[i]._methods;
      for (var prop in copyFrom) { copyTo[prop] = copyFrom[prop]; }

      // inlined object copy
      var copyTo = iface._interfaces; 
      var copyFrom = ifaces[i]._interfaces;
      for (var prop in copyFrom) { copyTo[prop] = copyFrom[prop]; }
    }
  }

  // process all properties in the descriptor
  for (var propertyName in descriptor) {
    // skip over the special properties
    if (propertyName == "$extends") {
      continue;
    }

    var value = descriptor[propertyName];

    if (value._$class_modifiedProperty) {
      iface[propertyName] = value.value;
    } else {
      iface._methods[propertyName] = iface._name;
    }
  }

  var namePath = name.split(".");
  var ifaceName = namePath.pop();
  // store the interface so it can be referenced by the specified name
  var scope = $class_library._globalObject;
  
  for (var i = 0, pathComponent; pathComponent = namePath[i]; ++i) {
    scope = scope[pathComponent];
  } 
  
  return scope[ifaceName] = iface;
}

//##############################################################################

$interface.prototype = {
  getName: function() {
    return this._name;
  },
  
  hasMethod: function(methodName) {
    return Boolean(this._methods[methodName]);
  },

  getMethods: function() {
    if (!this._methodsArray) {
      this._methodsArray = [];
      
      for (var methodName in this._methods) {
        this._methodsArray.push(methodName);
      }
    }
    
    return this._methodsArray;
  },

  applyInterface: function(obj) {
    var classInfo = obj instanceof $class ? obj : obj.$class;

    if (classInfo) {
      if (!classInfo.implementsInterface(this)) {
        this._applyInterface([classInfo]);
      }
    } else {
      if (obj.constructor == Function) {
        obj = obj.prototype;
      }
      
      if (!$class.implementationOf(obj, this)) {
        if (!obj._$class_interfaces) {
          obj._$class_interfaces = {};
        }

        for (var ifaceName in this._interfaces) {
          obj._$class_interfaces[ifaceName] = this._interfaces[ifaceName];
        }
      }
    }

    return obj;
  },

  _applyInterface: function(classes) {
    for (var i = 0, classInfo; classInfo = classes[i]; ++i) {
      for (var ifaceName in this._interfaces) {
        classInfo._interfaces[ifaceName] = this._interfaces[ifaceName];
      }

      this._applyInterface(classInfo._children);
    }
  },

  wrapObject: function(obj) {
    var result = {_obj: obj};

    for (var methodName in this._methods) {
      result[methodName] = $interface._createWrappedObjectMethod(methodName);
    }

    return this.applyInterface(result);
  },
  
  toString: function() {
    return "[$interface " + this._name + "]";
  }
};

$interface._createWrappedObjectMethod = function(methodName) {
  return function() {
    return this._obj[methodName].apply(this._obj, arguments);
  };
};

//##############################################################################

function $abstract(method) {
  // abstract modifier always ignored in release version
  return {};
}

function $static(value) {
  return {value: value, _$class_modifiedProperty: true};
}

function $final(method) {
  // final modifier always ignored in release version
  return method;
}

//##############################################################################

(function() {
  var nativeClassNames = [
    "Object",
    "Array",
    "String",
    "Number",
    "Boolean",
    "Function",
    "RegExp",
    "Date",

    "Error",
    "EvalError",
    "RangeError",
    "ReferenceError",
    "SyntaxError",
    "TypeError",
    "URIError"
  ];
  
  Object._$class_isTrivialCtor = true;
  
  for (var i = 0, className; className = nativeClassNames[i]; ++i) {
    $class.adapt(this[className], className, true);
  }
})();

$class.adapt($namespace, "$namespace");
$class.adapt($class,     "$class");
$class.adapt($interface, "$interface");

//##############################################################################

$class("Undefined", {
  Undefined: $final(function() {
    throw new Error("Attempted instantiation of the Undefined class.");
  })
});

$class("Null", {
  Null: $final(function() {
    throw new Error("Attempted instantiation of the Null class.");
  })
});

//##############################################################################
